import {
  ClassType, PropertyKeyType,
} from '@overflow/commons/core/type';
import {
  Class, Annotation, Field,
} from '@overflow/commons/core/reflect';

import {
  ValueAnnotation, InjectAnnotation, ResourceAnnotation,
} from '@overflow/commons/di/decorators';

import {
  InstanceNameType,
  InstanceFactoryType,
} from '@overflow/commons/di/type';


import {InstanceDefinitionRegistry} from './instance_definition_registry';
import {InstanceDefinition} from './instance_definition';

export class InstanceFactory implements InstanceDefinitionRegistry {
  private _singletonObjects: Map<InstanceNameType, any> = new Map();
  private _singletonFactories: Map<InstanceNameType, InstanceFactoryType> = new Map();
  private _instanceDefinitionMap: Map<InstanceNameType, InstanceDefinition> = new Map();
  private _instanceDefinitionNames: Set<InstanceNameType> = new Set();
  private _values: any;
  /**
   * getInstance
   */
  public getInstance(name: InstanceNameType, requiredType: ClassType, ...args: any[]): any | undefined {
    if (null === name) {
      name = requiredType.name;
    }

    let sharedInstance = this.getSingleton(name);
    if (undefined !== sharedInstance) {
      return sharedInstance;
    }

    let instanceDefinition = this.getInstanceDefinition(name);
    let clazzType = instanceDefinition.instanceClass;
    let clazz = Class.forClass(clazzType);
    if (undefined === clazz) {
      return undefined;
    }
    let ctor = clazz.getConstructor();
    let ctorParams = ctor.getParameters();
    let ctorArgs = [];
    if (0 < ctorParams.length) {
      //
    }

    let instance = ctor.newInstance(ctorArgs);

    this.applyInstance(clazzType, instance);

    return instance;
  }

  private getInstanceByName(requiredType: ClassType): any | undefined {
    return undefined;
  }

  private getInstanceByType(requiredType: ClassType): any | undefined {
    return undefined;
  }

  /**
   * applyInstance
   */
  public applyInstance(clazzType: ClassType, instance: any): void {
    let clazz: Class = Class.forClass(clazzType);

    let fields = clazz.getFields();

    fields.forEach((field: Field, key: PropertyKeyType, map: Map<PropertyKeyType, Field>): void => {
      let fieldAnnotations = field.getAnnotations();
      fieldAnnotations.forEach((value: Annotation, annonType: ClassType, annonMap: Map<ClassType, Annotation>): void => {
        switch (annonType) {
          case ValueAnnotation:
            {
            let annon = <ValueAnnotation>value;
            instance[field.getName()] = this.getValue(annon.attributes.value);
            }
            break;
          case InjectAnnotation:
            {
            let annon = <InjectAnnotation>value;
            instance[field.getName()] = this.getInstance(annon.attributes.name, field.getType());
            }
            break;
          case ResourceAnnotation:
            {
            let annon = <ResourceAnnotation>value;
            instance[field.getName()] = this.getSingleton(annon.attributes.name || field.getName());
            }
            break;
          default:
            break;
        }
      });

    });
  }

  private getSingleton(name: InstanceNameType): any | undefined {
    let singletonObject = this._singletonObjects.get(name);
  }

  /**
   * containsInstance
   */
  public containsInstance(name: InstanceNameType): boolean {
    return false;
  }

  public registerJSON(json: any): void {
    this._values = json;
  }

  /**
   * getValue
   */
  public getValue(keyPath: string): any {
    let paths = keyPath.split('.');
    let result = this._values;
    for (let i = 0, len = paths.length; i < len; i++) {
      result = result[paths[i]];
    }
    return result;
  }

  public registerSingleton(name: InstanceNameType, instanceType: any, instance: any): void {
    this._singletonObjects.set(name, instance);
  }

  public registerInstanceDefinition(name: InstanceNameType, instanceDefinition: InstanceDefinition): void {
    let oldInstanceDefinition = this._instanceDefinitionMap.get(name);
    if (undefined !== oldInstanceDefinition) {
      throw new Error(`InstanceDefinition[${name}] is already exist.`);
    }
    this._instanceDefinitionNames.add(name);
    this._instanceDefinitionMap.set(name, instanceDefinition);
  }

  public removeInstanceDefinition(name: InstanceNameType): void {
    if (!this._instanceDefinitionMap.has(name)) {
      throw new Error(`InstanceDefinition[${name}] is not exist.`);
    }
    this._instanceDefinitionNames.delete(name);
    this._instanceDefinitionMap.delete(name);
  }

  public getInstanceDefinition(name: InstanceNameType): InstanceDefinition | undefined {
    return this._instanceDefinitionMap.get(name);
  }

  public containsInstanceDefinition(name: InstanceNameType): boolean {
    return this._instanceDefinitionMap.has(name);
  }

  public getInstanceDefinitionNames(): Set<InstanceNameType> {
    return this._instanceDefinitionNames;
  }

  public getInstanceDefinitionCount(): number {
    return this._instanceDefinitionMap.size;
  }

  public isInstanceNameInUse(name: InstanceNameType): boolean {
    return false;
  }

}
